<?php
/**
 * Main test unit
 * 
 * USAGE:
 * - chdir to the current directory
 * - run on the command line : phpunit UmlParserTest
 * 
 * UmlParserTest compares the objects that PHP_UML_PHP_Parser has generated
 * against the serialized "correct" versions (stored in /providers).
 * It does it, first for each bug case (independantly), and then globally
 * (all cases at once), to discern errors implying the type resolution
 * (which works globally).
 * 
 * Beware : because PHPUnit relies on arrays to make its object reflection, and since there are
 * circular references between the EMOF program elements (because of the bidirectional relations),
 * UMLTest might warn you of an error in a "function", whereas the error is in a "class"
 * (or in a "class", whereas the problem is in a "package"). This behaviour remains to be fixed.
 * 
 * Also, note that if an error occurs in the one-by-one test, it is likely you'll
 * see it in the global test too.
 * 
 * PHP version 5
 * 
 * @category PHP
 * @package  PHP_UML::tests
 * @author   Baptiste Autin <ohlesbeauxjours@yahoo.fr>
 * @license  http://www.gnu.org/licenses/lgpl.html LGPL License 3
 * @version  SVN: <0.3>
 * @link     http://www.phpunit.de/
 * 
 */

error_reporting(E_ALL);

require_once 'PHPUnit/Framework.php';
require_once 'PHP/UML.php';

/**
 * Test unit class for the parser
 *
 * @category PHP
 * @package  PHP_UML::tests
 */
class UmlParserTest extends PHPUnit_Framework_TestCase
{
    const SUITE_DIR     = './suite/';
    const PROVIDERS_DIR = 'data-providers/';
    
    static public $IGNORED_DIR = array('.svn');
    
    /**
     * Provides the data for the "one by one" model check
     *
     * @return array
     */
    static public function providerModelOneByOne()
    {
        $t = new PHP_UML();
        $t->ignoredDirectories = self::$IGNORED_DIR;

        $data = array();
        foreach (new DirectoryIterator(self::SUITE_DIR) as $file) {
            if (!$file->isDot() && !$file->isDir()) {
                $filename = $file->getFilename();
                $t->parseFile(self::SUITE_DIR.$filename);
                $model  = unserialize(
                    file_get_contents(
                        self::SUITE_DIR.self::PROVIDERS_DIR.$filename.'.obj'
                    )
                );
                $data[] = array($model, $t->model, $filename);
            }
        }
        return $data;
    }

    /**
     * Provides the data for the "all at once" model check
     *
     * @return array
     */
    static public function providerModelGlobal()
    {
        $t = new PHP_UML();
        $t->ignoredDirectories = self::$IGNORED_DIR;

        $data = array();
        $t->parseDirectory(self::SUITE_DIR);
        $model  = unserialize(
            file_get_contents(
                self::SUITE_DIR.self::PROVIDERS_DIR.'global.obj'
            )
        );
        $data[] = array($model, $t->model, 'Global parsing');

        return $data;
    }
    
    /**
     * Checks the model of each bug case, one by one
     * 
     * @param mixed  $expected Expected element
     * @param mixed  $actual   Current element
     * @param string $filename Filename
     * 
     * @dataProvider providerModelOneByOne
     */
    public function testModelOneByOne($expected, $actual, $filename)
    {
        $this->_assertModel($expected->parameters, $actual->parameters, 'parameter', $filename);
        $this->_assertModel($expected->operations, $actual->operations, 'operation', $filename);
        $this->_assertModel($expected->classes, $actual->classes, 'class', $filename);
        $this->_assertModel($expected->interfaces, $actual->interfaces, 'interface', $filename);
        $this->_assertModel($expected->packages, $actual->packages, 'package', $filename);
        $this->_assertModel($expected->files, $actual->files, 'file', $filename);
    }
 
    /**
     * Checks the model globally (all bug cases at once)
     * 
     * @param mixed  $expected Expected element
     * @param mixed  $actual   Current element
     * @param string $filename Filename
     * 
     * @dataProvider providerModelGlobal
     */
    public function testModelGlobal($expected, $actual, $filename)
    {
        $this->_assertModel($expected->parameters, $actual->parameters, 'parameter', $filename);
        $this->_assertModel($expected->operations, $actual->operations, 'operation', $filename);
        $this->_assertModel($expected->classes, $actual->classes, 'class', $filename);
        $this->_assertModel($expected->interfaces, $actual->interfaces, 'interface', $filename);
        $this->_assertModel($expected->packages, $actual->packages, 'package', $filename);
        //$this->_assertModel($expected->files, $actual->files, 'file', $filename);
    }

    private function _assertModel(PHP_UML_Metamodel_Sequence $actual, PHP_UML_Metamodel_Sequence $expected, $type, $filename)
    {
        $actIter = $actual->getIterator();
        $expIter = $expected->getIterator();
        $this->assertEquals(count($actIter), count($expIter), $type.' : different numbers');
        foreach ($expIter as $e) {
            $objE = $expIter->current();
            $objA = $actIter->current();
            $this->_assertEMOFEqual($objE, $objA, $type, $filename);
            $actIter->next();
        }
    }
    
    private function _assertEMOFEqual(PHP_UML_Metamodel_NamedElement $actual, PHP_UML_Metamodel_NamedElement $expected, $type, $filename)
    {
        $this->assertEquals(
            get_class($actual),
            get_class($expected),
            get_class($actual).' and '.get_class($expected).': different classes'
        );
        $a = (array) $actual;
        $b = (array) $expected;
        $this->assertEquals(count($a), count($b), $type.' : different numbers');
        foreach ($a as $key => $value) {
            if (is_array($value)) {
                $bb = $b[$key];
                foreach ($value as $k => $v) {
                    $this->assertEquals($v, $bb[$k], 'In: '.$filename.'/'.$type.'/'.$key.'/'.$k, 0, 2);
                }
            }
            else {
                $this->assertEquals($b[$key], $value, 'In: '.$filename.'/'.$type.'/'.$key, 0, 1);
            }
        }
    }
}

?>